"""This module implements pre-questions' progressions.

Pre questions are chord progressions or notes played before the question is
played, so to affirmate the sound of the question's key.

For example a common cadence is chords I-IV-V-I from the diatonic scale, which
in a key of `C` is `CM-FM-GM-CM` and in a key of `A` is `AM-DM-EM-AM`.

Pre-question methods should be decorated with `register_prequestion_method`
decorator, so that they will be registered as a valid pre-question method.
"""

from .sequence import Sequence

from functools import wraps

PREQUESTION_METHODS = {}
# http://stackoverflow.com/questions/5910703/howto-get-all-methods-of-a-pyt\
# hon-class-with-given-decorator
# http://stackoverflow.com/questions/5707589/calling-functions-by-array-ind\
# ex-in-python/5707605#5707605


def register_prequestion_method(f, *args, **kwargs):
    """Decorator for prequestion method functions.

    Functions decorated with this decorator will be registered in the
    `PREQUESTION_METHODS` global dict.
    """

    @wraps(f)
    def decorator(*args, **kwargs):
        return f(*args, **kwargs)

    PREQUESTION_METHODS.update({f.__name__: f})

    return decorator


class PreQuestion:

    def __init__(self, method, question):
        """This class implements methods for different types of pre-question
        progressions.

        Args:
            method (str): The method used in the pre question.
            question (obj): Question object from which to generate the
            pre-question sequence.
        """

        self.METHOD = PREQUESTION_METHODS[method]
        self.question = question

    def __call__(self, *args, **kwargs):
        """Calls the resolution method and pass arguments to it.

        Returns a `birdears.Sequence` object with the pre-question generated by
        the method.
        """
        return self.METHOD(question=self.question, *args, **kwargs)


@register_prequestion_method
def none(question, *args, **kwargs):
    """Pre-question method that return an empty sequence with no delay.
    Args:
        question (obj): Question object from which to generate the
            pre-question sequence. (this is provided by the `Resolution` class
            when it is `__call__`ed)
    """
    return Sequence(elements=[], duration=0, delay=0, pos_delay=0)


@register_prequestion_method
def tonic_only(question, *args, **kwargs):
    """Pre-question method that only play's the question tonic note before the
    question.

    Args:
        question (object): Question object from which to generate the
            pre-question sequence. (this is provided by the `Resolution` class
            when it is `__call__`ed)
    """

    tonic_pitch = question.tonic_pitch

    duration = question.durations['preq']['duration']
    delay = question.durations['preq']['delay']
    pos_delay = question.durations['preq']['pos_delay']

    sequence = Sequence(elements=[tonic_pitch], duration=duration,
                        delay=delay, pos_delay=pos_delay)

    return sequence


@register_prequestion_method
def progression_i_iv_v_i(question, *args, **kwargs):
    """Pre-question method that play's a chord progression with triad chords
    built on the grades I, IV, V the I of the question key.

    Args:
        question (obj): Question object from which to generate the
            pre-question sequence. (this is provided by the `Resolution` class
            when it is `__call__`ed)
    """

    # tonic_pitch = question.tonic_pitch
    tonic_str = question.tonic_str
    mode = question.mode

    duration = question.durations['preq']['duration']
    delay = question.durations['preq']['delay']
    pos_delay = question.durations['preq']['pos_delay']

    degrees = [1, 4, 5, 1]

    sequence = Sequence(duration=duration, delay=delay,
                        pos_delay=pos_delay)

    sequence.make_chord_progression(tonic=tonic_str, mode=mode,
                                    degrees=degrees)

    return sequence



from . import INTERVALS


def get_interval_by_semitones(semitones):
    # TODO: PLEASE WRITE ME
    pass


class Interval(dict):
    """This class represents the interval between two pitches..

    Attributes:
        tonic_octave (int): Scientific octave for the tonic. For example, if
            the tonic is a 'C4' then `tonic_octave` is 4.
        interval octave (int): Scientific octave for the interval. For example,
            if the interval is a 'G5' then `tonic_octave` is 5.
        chromatic_offset (int): The offset in semitones inside one octave.
            Relative semitones to tonic.
        note_and_octave (str): Note and octave of the interval, for example, if
            the interval is G5 the note name is 'G5'.
        note_name (str): The note name of the interval, for example, if the
            interval is G5 then the name is 'G'.
        semitones (int): Semitones from tonic to octave. If tonic is C4 and
            interval is G5 the number of semitones is 19.
        is_chromatic (bool): If the current interval is chromatic (True) or if
            it exists in the diatonic scale which key is tonic.
        is_descending (bool): If the interval has a descending direction, ie.,
            has a lower pitch than the tonic.
        diatonic_index (int): If the interval is chromatic, this will be the
            nearest diatonic interval in the direction of the resolution
            (closest tonic.) From II to IV degrees, it is the ditonic interval
            before; from V to VII it is the diatonic interval after.
        distance (dict): A dictionary which the distance from tonic to
            interval, for example, if tonic is C4 and interval is G5::
                {
                    'octaves': 1,
                    'semitones': 7
                }
        data (tuple): A tuple representing the interval data in the form of
            (semitones, short_name, long_name), for example::
                (19, 'P12', 'Perfect Twelfth')
    """

    def __init__(self, pitch_a, pitch_b):
        """Measures the musical interval from pitch_a to pitch_b.

        Args:
            pitch_a (str): First `Pitch` object to be measured.
            pitch_b (str): Second `Pitch` object to be measured.
        """

        descending = True if int(pitch_b) < int(pitch_a) else False

        semitones = abs(int(pitch_b) - int(pitch_a))

        self.update({
            'tonic_octave': pitch_a.octave,
            'tonic_note_and_octave': str(pitch_a),
            'interval_octave': pitch_b.octave,
            'chromatic_offset': pitch_b.pitch_class,
            'note_and_octave': str(pitch_b),
            'note_name': str(pitch_b.note),
            'note_octave': pitch_b.octave,
            'semitones': semitones,
            'is_descending': descending,
            'distance': {'octaves': int(semitones / 12),
                         'semitones': int(semitones % 12)},
            'data': INTERVALS[semitones],
        })



from . import CHROMATIC_SHARP
from . import CHROMATIC_FLAT

from .exception import InvalidNote
from .exception import InvalidOctave
from .exception import InvalidPitch

# pitch_numeric_value = (PITCH_CLASS) + (OCTAVE * 12)
# eg.: C4 == (0) + (4*12), == 60


def get_pitch_class(note):
    if note in CHROMATIC_SHARP:
        pitch_class = CHROMATIC_SHARP.index(note)
    elif note in CHROMATIC_FLAT:
        pitch_class = CHROMATIC_FLAT.index(note)
    else:
        raise InvalidNote

    return pitch_class


def get_pitch_number(note, octave):
    pitch_number = get_pitch_class(note) + (octave * 12)
    return pitch_number


def get_pitch_by_number(numeric, accident='sharp'):

    octave, pitch_class = divmod(numeric, 12)

    note = CHROMATIC_SHARP[pitch_class] if accident == 'sharp' \
        else CHROMATIC_FLAT[pitch_class]

    pitch = Pitch(note=note, octave=octave)

    return pitch


class Note:

    def __init__(self, note='C'):
        if note in CHROMATIC_SHARP or note in CHROMATIC_FLAT:
            self.note = note
        else:
            raise InvalidNote

    # https://en.wikipedia.org/wiki/Pitch_class
    @property
    def pitch_class(self):
        if self.note in CHROMATIC_SHARP:
            value = CHROMATIC_SHARP.index(self.note)
        else:
            value = CHROMATIC_FLAT.index(self.note)

        return value

    def __eq__(self, compare):
        # TODO: think a way to compare pitchs vs strings vs notes

        if type(compare) == str and str(self) == compare:
            return True

        elif type(compare) == Note and int(self) == int(compare):
            return True

        elif type(compare) == Pitch and int(self) == int(compare):
            return True

        return False

    def __int__(self):
        return self.pitch_class

    def __str__(self):
        return str(self.note)

    def __repr__(self):
        return "<Note '{note}'>".format(note=self.note)


# https://en.wikipedia.org/wiki/Scientific_pitch_notation
class Pitch(Note):

    duration = None
    delay = None

    def __init__(self, note='C', octave=4):
        super(Pitch, self).__init__(note=note)

        if octave >= 0 and octave <= 9:
            self.octave = octave
        else:
            raise InvalidOctave

    @property
    def pitch_number(self):
        value = self.pitch_class + (self.octave * 12)
        return value

    # FIXME: maybe move this somewhere else
    def get_pitch_by_number(self, numeric, accident='sharp'):
        octave, pitch_class = divmod(numeric, 12)

        note = CHROMATIC_SHARP[pitch_class] if accident == 'sharp' \
            else CHROMATIC_FLAT[pitch_class]

        pitch = Pitch(note=note, octave=octave)

        return pitch

    def distance(self, other):
        if type(other) == Pitch:
            return self.pitch_number - int(other)

    def __eq__(self, compare):

        if type(compare) == int:
            result = self.pitch_number == compare

        elif type(compare) == Pitch:
            result = self.pitch_number == int(compare)

        elif type(compare) == Note:
            result = self.pitch_class == int(compare)

        else:
            raise Exception('Invalid operand for addition.')

        return result

    def __str__(self):
        return "{note}{octave}".format(note=self.note, octave=self.octave)

    def __int__(self):
        return self.pitch_number

    def __repr__(self):
        return "<Pitch '{note}{octave}' ({numeric})>" \
            .format(note=self.note, octave=self.octave,
                    numeric=self.pitch_number)

    def __add__(self, other):
        if type(other) == int:
            pitch_number = self.pitch_number + other
            pitch = Pitch().get_pitch_by_number(pitch_number)
        else:
            raise Exception('Invalid operand for addition.')
        return pitch

    def __iadd__(self, other):
        if type(other) == int:
            pitch_number = self.pitch_number + other
            pitch = Pitch().get_pitch_by_number(pitch_number)
        else:
            raise Exception('Invalid operand for addition.')

        return pitch

    def __sub__(self, other):
        if type(other) == int:
            pitch_number = self.pitch_number - other
            pitch = Pitch().get_pitch_by_number(pitch_number)
        else:
            raise Exception('Invalid operand for subtraction.')

        return pitch

    def __isub__(self, other):
        if type(other) == int:
            pitch_number = self.pitch_number - other
            pitch = Pitch().get_pitch_by_number(pitch_number)
        else:
            raise Exception('Invalid operand for subtraction.')

        return pitch


class Chord(list):

    duration = None
    delay = None

    def __init__(self, iterable):
        if not all(isinstance(element, Pitch) for element in iterable):
            raise InvalidPitch
        # else:
        super(Chord, self).__init__(iterable)

    def append(self, obj):
        if not isinstance(obj, Pitch):
            raise InvalidPitch
        # else:
        super(Chord, self).append(obj)

    def extend(self, iterable):
        if not all(isinstance(element, Pitch) for element in iterable):
            raise InvalidPitch
        # else:
        super(Chord, self).extend(iterable)



from . import MAX_SEMITONES_RESOLVE_BELOW

from .interval import Interval

from .scale import DiatonicScale

from .sequence import Sequence

from .note_and_pitch import get_pitch_by_number
from .note_and_pitch import Chord

from functools import wraps

RESOLUTION_METHODS = {}
# http://stackoverflow.com/questions/5910703/howto-get-all-methods-of-a-pyt\
# hon-class-with-given-decorator
# http://stackoverflow.com/questions/5707589/calling-functions-by-array-ind\
# ex-in-python/5707605#5707605


def register_resolution_method(f, *args, **kwargs):
    """Decorator for resolution method functions.

    Functions decorated with this decorator will be registered in the
    `RESOLUTION_METHODS` global dict.
    """

    @wraps(f)
    def decorator(*args, **kwargs):
        return f(*args, **kwargs)

    RESOLUTION_METHODS.update({f.__name__: f})

    return decorator


class Resolution:
    """This class implements methods for different types of question
    resolutions.

    A resolution is an answer to a question. It aims to create a mnemonic on
    how the inverval resvolver to the tonic.
    """

    def __init__(self, method, question):
        """Inits the resolution class.

        Args:
            method (str): The method used in the resolution.
            question (obj): Question object from which to generate the
            resolution sequence.
        """

        self.METHOD = RESOLUTION_METHODS[method]
        self.question = question

    def __call__(self, *args, **kwargs):
        """Calls the resolution method and pass arguments to it.

        Returns a `birdears.Sequence` object with the resolution generated by
        the.method.
        """
        return self.METHOD(question=self.question, *args, **kwargs)


@register_resolution_method
def nearest_tonic(question):
    """Resolution method that resolve the intervals to their nearest tonics.

    Args:
        question (obj): Question object from which to generate the
            resolution sequence. (this is provided by the `Prequestion` class
            when it is `__call__`ed)
    """

    tonic_pitch = question.tonic_pitch

    # if hasattr(question, 'random_pitch'):
    #    random_pitch = tuple(question.random_pitch)
    # else:
    #    random_pitch = tuple(question.random_pitches)
    #
    # TODO:
    if hasattr(question, 'random_pitches'):
        raise Exception('NEAREST_TONIC FOR MULTIPLE PITCHES IS STILL TO BE'
                        'IMPLEMENTED')
    random_pitch = question.random_pitch

    duration = question.durations['resol']['duration']
    delay = question.durations['resol']['delay']
    pos_delay = question.durations['resol']['pos_delay']

    # this function will receive: tonic, scale and random_pitch (which may be
    # chromatic, ie., not in `scale`)

    semitones = (int(random_pitch) - int(tonic_pitch)) % 12

    scale_random_pitch = question.diatonic_scale

    direction = -1 if (semitones <= MAX_SEMITONES_RESOLVE_BELOW) else +1
    # invert `direction` signal if descending:
    pitch_direction = (direction * -1) if question.is_descending else direction

    resolution = []

    if random_pitch not in scale_random_pitch:  # random_pitch is chromatic
        resolution.append(random_pitch)
        # if this note is chromatic then the
        # next ones above or below are in the diatonic for sure, so we
        # add one semitone and we will have the next diatonic degree:
        nearest_diatonic_pitch = \
            get_pitch_by_number(int(random_pitch) + direction)
    else:
        nearest_diatonic_pitch = random_pitch  # random_pitch is diatonic

    resolve_mask = 0 if semitones <= MAX_SEMITONES_RESOLVE_BELOW else 12

    nearest_tonic_pitch = get_pitch_by_number(int(random_pitch) +
                                              (resolve_mask-semitones))

    nearest_tonic_index = scale_random_pitch.index(nearest_tonic_pitch)

    nearest_diatonic_pitch_index = \
        scale_random_pitch.index(nearest_diatonic_pitch)

    random_pitch_index = nearest_diatonic_pitch_index

    ohslice = slice(min(nearest_tonic_index, random_pitch_index),
                    max(nearest_tonic_index, nearest_diatonic_pitch_index)+1)

    resolution = scale_random_pitch[ohslice][::pitch_direction]

    # is it chromatic?
    if random_pitch not in scale_random_pitch:
        resolution.insert(0, random_pitch)

    if len(resolution) == 1:
        resolution.append(tonic_pitch)

    if question.is_harmonic:
        resolution_harmonic = []

        for item in resolution:
            resolution_harmonic.append(Chord([tonic_pitch, item]))

        resolution = resolution_harmonic

    # resolution
    return Sequence(elements=resolution, duration=duration, delay=delay,
                    pos_delay=pos_delay)


@register_resolution_method
# FIXME : it should both play preq and question
def repeat_only(question):
    """Resolution method that only repeats the sequence elements with given
    durations.

    Args:
        question (obj): Question object from which to generate the
            resolution sequence. (this is provided by the `Prequestion` class
            when it is `__call__`ed)
    """

    elements = list()
    if hasattr(question, 'random_pitches'):
        elements.extend(question.random_pitches)
    else:
        elements.append(question.random_pitch)
        elements.append(question.tonic_pitch)

    duration = question.durations['resol']['duration']
    delay = question.durations['resol']['delay']
    pos_delay = question.durations['resol']['pos_delay']

    sequence_list = Sequence(elements, duration=duration, delay=delay,
                             pos_delay=pos_delay)

    return sequence_list



from random import randrange
from random import choice

from . import KEYBOARD_INDICES
from . import CHROMATIC_TYPE
from . import KEYS

from .interval import Interval

from .note_and_pitch import Pitch

from functools import wraps

QUESTION_CLASSES = {}


def register_question_class(function, *args, **kwargs):
    """Decorator for question classes.

    Classes decorated with this decorator will be registered in the
    `QUESTION_CLASSES` global.
    """

    @wraps(function)
    def decorator(*args, **kwargs):
        return function(*args, **kwargs)

    QUESTION_CLASSES.update({function.__name__: function})

    return decorator


# values for valid_semitones list can be Interval objects or int's (semitones)
def get_valid_pitches(scale, valid_semitones=CHROMATIC_TYPE):
    tonic_pitch = scale[0]

    valid_scale = list()

    if isinstance(valid_semitones, tuple):
        pass
    elif isinstance(valid_semitones, list):
        pass
    elif isinstance(valid_semitones, str):
        valid_list = valid_semitones.replace(' ', '').split(',')
        valid_semitones = tuple(map(lambda x: int(x)-1, valid_list))
    else:
        raise Exception('Incorrect type for valid_semitones')

    for pitch in scale:

        chromatic_offset = \
            abs(int(tonic_pitch) - int(pitch))

        if chromatic_offset in valid_semitones:
            valid_scale.append(pitch)

    return valid_scale


class QuestionBase:
    """
    Base Class to be subclassed for Question classes.

    This class implements attributes and routines to be used in Question
    subclasses.
    """

    def __init__(self, mode='major', tonic='C', octave=4, descending=False,
                 chromatic=False, n_octaves=1, valid_intervals=CHROMATIC_TYPE,
                 user_durations=None, prequestion_method=None,
                 resolution_method=None, default_durations=None,
                 *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string represnting the mode of the question.
                Eg., 'major' or 'minor'
            tonic (str): A string representing the tonic of the
                question, eg.: 'C'; if omitted, it will be selected
                randomly.
            octave (int): A scienfic octave notation, for example,
                4 for 'C4'; if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending,
                ie., intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not
                (False) chromatic intervals, ie., intervals not in the
                diatonic scale of tonic/mode.
            n_octaves (int): Maximum numbr of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
            user_durations (dict): Dictionary with the default durations for
                each type of sequence. This is provided by the subclasses.
        """

        self.mode = mode

        self.is_descending = descending
        self.is_chromatic = chromatic

        # self.octave = octave if octave else randrange(3, 5)
        if not octave:
            octave = randrange(3, 5)
        elif isinstance(octave, list):
            octave = choice(octave)
        elif isinstance(octave, tuple) and len(octave) == 2:
            octave = randrange(*octave)

        # TODO: raise exceptions in case octave/n_octaves are invalid or
        #       extrapolate each other

        self.octave = octave

        self.n_octaves = n_octaves

        direction = 'descending' if descending else 'ascending'

        # FIXME: maybe this should go to __main__
        self.keyboard_index = \
            tuple(KEYBOARD_INDICES['chromatic'][direction][self.mode])

        if isinstance(tonic, list) or isinstance(tonic, tuple):
            tonic = choice(tonic)
        elif isinstance(tonic, str) and ',' in tonic:
            tonic.replace(' ', '')
            tonic = choice(tonic.split(','))
        elif isinstance(tonic, str) and 'R' in tonic:
            tonic = choice(KEYS)

        self.tonic_pitch = Pitch(note=tonic, octave=self.octave)
        self.tonic_str = str(self.tonic_pitch.note)
        self.tonic_pitch_str = str(self.tonic_pitch)

        self.durations = default_durations
        if user_durations:
            ud_index = {
                0: ('preq', 'duration'),
                1: ('preq', 'delay'),
                2: ('preq', 'pos_delay'),
                3: ('quest', 'duration'),
                4: ('quest', 'delay'),
                5: ('quest', 'pos_delay'),
                6: ('resol', 'duration'),
                7: ('resol', 'delay'),
                8: ('resol', 'pos_delay'),
            }
            ud = user_durations.split(',')
            if len(ud) == len(ud_index):
                for idx, v in ud_index.items():
                    cur_duration = ud[idx].strip()
                    if cur_duration != 'n':
                        self.durations[v[0]][v[1]] = float(cur_duration)

        self.prequestion_method = prequestion_method
        self.resolution_method = resolution_method

    def make_question(self):
        """This method should be overwritten by the question subclasses.
        """

        pass

    def make_resolution(self):
        """This method should be overwritten by the question subclasses.
        """

        pass

    def play_question(self):
        """This method should be overwritten by the question subclasses.
        """

        pass

    def check_question(self):
        """This method should be overwritten by the question subclasses.
        """

        pass



from .. import CHROMATIC_TYPE

from ..questionbase import QuestionBase
from ..questionbase import get_valid_pitches

from ..interval import Interval

from ..scale import DiatonicScale
from ..scale import ChromaticScale

from ..sequence import Sequence
from ..resolution import Resolution
from ..prequestion import PreQuestion

from ..interfaces.commandline import center_text

from random import choice
from random import sample


class InstrumentalDictationQuestion(QuestionBase):
    """Implements an instrumental dictation test.
    """

    def __init__(self, mode='major', wait_time=11, n_repeats=1,
                 max_intervals=3, n_notes=4, tonic='C', octave=4,
                 descending=False, chromatic=False, n_octaves=1,
                 valid_intervals=CHROMATIC_TYPE, user_durations=None,
                 prequestion_method='progression_i_iv_v_i',
                 resolution_method='repeat_only', *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string representing the mode of the question.
                Eg., 'major' or 'minor'.
            wait_time (float): Wait time in seconds for the next question or
                repeat.
            n_repeats (int): Number of times the same dictation will be
                repeated before the end of the exercise.
            max_intervals (int): The maximum number of random intervals the
                question will have.
            n_notes (int): The number of notes the melodic dictation will have.
            tonic (str): A string representing the tonic of the question,
                eg.: 'C'; if omitted, it will be selected randomly.
            octave (int): A scienfic octave notation, for example, 4 for 'C4';
                if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending, ie.,
                intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not (False)
                chromatic intervals, ie., intervals not in the diatonic scale
                of tonic/mode.
            n_octaves (int): Maximum number of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise. Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
        """

        default_durations = {
            'preq': {'duration': 2, 'delay': 0.5, 'pos_delay': 1},
            'quest': {'duration': 2, 'delay': 0.5, 'pos_delay': 0},
            'resol': {'duration': 2.5, 'delay': 0.5, 'pos_delay': 1}
        }

        super(InstrumentalDictationQuestion, self).\
            __init__(mode=mode, tonic=tonic, octave=octave,
                     descending=descending, chromatic=chromatic,
                     n_octaves=n_octaves, valid_intervals=valid_intervals,
                     user_durations=user_durations,
                     prequestion_method=prequestion_method,
                     resolution_method=resolution_method,
                     default_durations=default_durations, *args, **kwargs)

        self.is_harmonic = False

        self.wait_time = wait_time
        self.n_repeats = n_repeats

        if not chromatic:
            self.scale = DiatonicScale(tonic=tonic, mode=mode, octave=octave,
                                       n_octaves=n_octaves,
                                       descending=descending)
        else:
            self.scale = ChromaticScale(tonic=tonic, octave=octave,
                                        n_octaves=n_octaves,
                                        descending=descending)

        self.valid_pitches = get_valid_pitches(self.scale, valid_intervals)

        self.random_pitches = [choice(self.valid_pitches) for note
                               in range(n_notes)]

        self.pre_question =\
            self.make_pre_question(method=prequestion_method)
        self.question = self.make_question()
        self.resolution = self.make_resolution(method=resolution_method)

    def make_pre_question(self, method):
        prequestion = PreQuestion(method=method, question=self)

        return prequestion()

    def make_question(self):
        return Sequence(self.random_pitches, **self.durations['quest'])

    def make_resolution(self, method):

        resolve = Resolution(method=method, question=self)

        resolution = resolve()

        return resolution

    def play_question(self):

        for r in range(self.n_repeats):
            self.pre_question.play()
            self.question.play()

            for i in range(self.wait_time):
                time_left = str(self.wait_time - i).rjust(3)
                text = '{} seconds remaining...'.format(time_left)
                print(center_text(text, sep=False), end='')
                self.question._wait(1)

    def check_question(self):
        """Checks whether the given answer is correct.

        This currently doesn't applies to instrumental dictation questions.
        """

        intervals = " ".join([Interval(self.tonic_pitch, pitch)['data'][1]
                             for pitch in self.random_pitches]).center(7)

        notes = " ".join([str(pitch) for pitch in self.random_pitches]) \
                .center(7)

        correct_response_str = """\
The intervals and notes of this question:

{intervals}
{notes}
""".format(**dict(intervals=intervals, notes=notes))
        response = {
            'correct_semitones': self.random_pitches,
            'correct_response_str': correct_response_str
        }

        return response



from .. import INTERVALS

from ..logger import log_event

from ..questionbase import QuestionBase
from ..questionbase import get_valid_pitches

from ..scale import DiatonicScale
from ..scale import ChromaticScale

from ..interval import Interval

from .. import KEYBOARD_INDICES
from .. import CHROMATIC_SHARP
from .. import CHROMATIC_FLAT
from .. import CHROMATIC_TYPE

from ..sequence import Sequence
from ..resolution import Resolution
from ..prequestion import PreQuestion

from random import choice


class NoteNameQuestion(QuestionBase):
    """Implements a Note Name test.
    """

    @log_event
    def __init__(self, mode='major', tonic='C', octave=4, descending=False,
                 chromatic=False, n_octaves=1, valid_intervals=CHROMATIC_TYPE,
                 user_durations=None, prequestion_method='tonic_only',
                 resolution_method='nearest_tonic', *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string representing the mode of the question.
                Eg., 'major' or 'minor'
            tonic (str): A string representing the tonic of the question,
                eg.: 'C'; if omitted, it will be selected randomly.
            octave (int): A scienfic octave notation, for example, 4 for 'C4';
                if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending, ie.,
                intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not (False)
                chromatic intervals, ie., intervals not in the diatonic scale
                of tonic/mode.
            n_octaves (int): Maximum number of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise. Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
        """

        default_durations = {
            'preq': {'duration': 2, 'delay': 0.5, 'pos_delay': 1},
            'quest': {'duration': 2, 'delay': 0.5, 'pos_delay': 0},
            'resol': {'duration': 2.5, 'delay': 0.5, 'pos_delay': 1}
        }

        super(NoteNameQuestion, self).\
            __init__(mode=mode, tonic=tonic, octave=octave,
                     descending=descending, chromatic=chromatic,
                     n_octaves=n_octaves, valid_intervals=valid_intervals,
                     user_durations=user_durations,
                     prequestion_method=prequestion_method,
                     resolution_method=resolution_method,
                     default_durations=default_durations, *args, **kwargs)

        self.is_harmonic = False

        if not chromatic:
            self.scale = DiatonicScale(tonic=self.tonic_str, mode=mode,
                                       octave=self.octave,
                                       descending=descending,
                                       n_octaves=n_octaves)
        else:
            self.scale = ChromaticScale(tonic=self.tonic_str,
                                        octave=self.octave,
                                        descending=descending,
                                        n_octaves=n_octaves)

        self.diatonic_scale = DiatonicScale(tonic=self.tonic_str, mode=mode,
                                            octave=self.octave,
                                            descending=descending,
                                            n_octaves=n_octaves)

        self.chromatic_scale = ChromaticScale(tonic=self.tonic_str,
                                              octave=self.octave,
                                              descending=descending,
                                              n_octaves=n_octaves)

        self.valid_pitches = get_valid_pitches(self.scale, valid_intervals)
        self.random_pitch = choice(self.valid_pitches)

        self.interval = Interval(self.tonic_pitch, self.random_pitch)

        self.pre_question = self.make_pre_question(method=prequestion_method)
        self.question = self.make_question()
        self.resolution = self.make_resolution(method=resolution_method)

    def make_pre_question(self, method):

        prequestion = PreQuestion(method=method, question=self)

        return prequestion()

    def make_question(self):

        question = Sequence([self.random_pitch], **self.durations['quest'])

        return question

    def make_resolution(self, method):

        resolve = Resolution(method=method, question=self)
        resolution = resolve()

        return resolution

    def play_question(self):
        # Other threads can call a thread’s join() method. This blocks the
        # calling thread until the thread whose join() method is called is
        # terminated.
        # https://docs.python.org/3/library/threading.html#thread-objects

        self.pre_question.play()
        self.question.play()

    def play_resolution(self):

        thread = self.resolution.play()
        thread.join()

    def check_question(self, user_input_char):
        """Checks whether the given answer is correct.
        """

        global INTERVALS

        from ..scale import ChromaticScale
        c_chromatic = ChromaticScale(tonic='C', n_octaves=2)
        # question_tone_chromatic = ChromaticScale(tonic=self.tonic_str,
        #                                         octave=self.octave,
        #                                         n_octaves = self.n_octaves)

        keyboard_index = \
            KEYBOARD_INDICES['chromatic']['ascending']['major']

        user_semitones = keyboard_index.index(user_input_char[0])

        user_pitch = c_chromatic[user_semitones]
        user_note = user_pitch.note

        correct_semitones = abs(int(self.tonic_pitch) - int(self.random_pitch))
        correct_note = self.random_pitch.note

        if user_note in CHROMATIC_SHARP:
            user_semitones = CHROMATIC_SHARP.index(user_note)
        elif user_note in CHROMATIC_FLAT:
            user_semitones = CHROMATIC_SHARP.index(user_note)

        if correct_note in CHROMATIC_SHARP:
            correct_semitones = CHROMATIC_SHARP.index(user_note)
        elif correct_note in CHROMATIC_FLAT:
            correct_semitones = CHROMATIC_FLAT.index(correct_note)

        signal = '✓' if user_semitones == correct_semitones else 'x'  # u2713

        extra_response_str = """\
       “{}”
user {} “{}”
{} semitones
""".format(correct_note, signal, user_note, self.interval['semitones'])

        response = dict(
            is_correct=False,
            # user_interval=user_interval,
            user_note=user_note,
            # correct_interval=correct_interval,
            correct_note=correct_note,
            user_response_str=user_note,
            correct_response_str=correct_note,
            extra_response_str=extra_response_str,
        )

        if user_note == correct_note:
            response.update({'is_correct': True})

        else:
            response.update({'is_correct': False})

        return response



from . import DIATONIC_FORMS

from .note_and_pitch import Pitch
from .note_and_pitch import Chord
from .note_and_pitch import get_pitch_by_number

from itertools import cycle

# https://docs.python.org/3/reference/datamodel.html#emulating-container-types


class ScaleBase(list):
    def __init__(self):
        pass


class DiatonicScale(ScaleBase):
    """Builds a musical diatonic scale.

    Attributes:
        scale (array_type): The array of notes representing the scale.
    """

    def __init__(self, tonic='C', mode='major', octave=4, n_octaves=1,
                 descending=False, dont_repeat_tonic=False):
        """Returns a diatonic scale from tonic and mode.

        Args:
            tonic (str): The note which the scale will be built upon.
            mode (str): The mode the scale will be built upon.
                ('major' or 'minor')
            octave (int): The scientific octave the scale will be built upon.
            n_octaves (int): The number of octaves the scale will contain.
            descending (bool): Whether the scale is descending.
            dont_repeat_tonic (bool): Whether to skip appending the last
                note (octave) to the scale.
        """

        super(DiatonicScale, self).__init__()

        self.tonic = Pitch(note=tonic, octave=octave)
        self.mode = mode
        self.direction = "Ascending" if not descending else "Descending"

        diatonic_mode = DIATONIC_FORMS[mode]

        form_length = len(diatonic_mode)

        direction = +1 if not descending else -1
        repeat_tonic = 0 - dont_repeat_tonic  # 0 (repeat) or -1
        if descending:
            diatonic_mode = diatonic_mode[::-1]

        diatonic_loop = cycle(diatonic_mode)

        scale = list()

        self.append(self.tonic)

        pitch_num = int(self[0])

        for i in range((form_length * n_octaves) + repeat_tonic):
            step = next(diatonic_loop)

            pitch_num += step * direction

            pitch = get_pitch_by_number(pitch_num)
            scale.append(pitch)

        self.extend(scale)

    # FIXME: aybe ake this a function
    def get_triad(self, index=0, degree=None):
        """Returns an array with notes from a scale's triad.

        Args:
            index (int): triad index (eg.: 0 for 1st degree triad.)
            degree (int): Degree of the scale. If provided, overrides the
                `index` argument. (eg.: `1` for the 1st degree triad.)
        Returns:
            An array with three pitches, one for each note of the triad.
        """

        tonic = self.tonic
        mode = self.mode

        diatonic = DiatonicScale(tonic=tonic.note, mode=mode,
                                 octave=tonic.octave, n_octaves=2,
                                 descending=False, dont_repeat_tonic=False)

        if degree:
            index = degree - 1

        form = [0, 2, 4]

        triad = [diatonic[index+note] for note in form]

        chord = Chord(triad)

        return chord

    def __repr__(self):

        repr = "<DiatonicScale {tonic} {mode} {direction} {first}-{to} " \
               "({octaves} octaves)>" \
                   .format(tonic=str(self[0].note),
                           mode=self.mode.capitalize(),
                           direction=self.direction.capitalize(),
                           first=str(self[0]),
                           to=str(self[-1]), octaves=int(len(self)/8))
        return repr

    def __str__(self):
        return str(list(self))


class ChromaticScale(ScaleBase):
    """Builds a musical chromatic scale.

    Attributes:
        scale (array_type): The array of notes representing the scale.
    """

    def __init__(self, tonic='C', octave=4, n_octaves=1, descending=False,
                 dont_repeat_tonic=False):
        """Returns a chromatic scale from tonic.

        Args:
            tonic (str): The note which the scale will be built upon.
            octave (int): The scientific octave the scale will be built upon.
            n_octaves (int): The number of octaves the scale will contain.
            descending (bool): Whether the scale is descending.
            dont_repeat_tonic (bool): Whether to skip appending the last
                note (octave) to the scale.
        """

        super(ChromaticScale, self).__init__()

        # global CHROMATIC_SHARP, CHROMATIC_FLAT

        self.tonic = Pitch(tonic, octave)

        direction = +1 if not descending else -1

        tonic_pitch_num = int(self.tonic)
        repeat_tonic = not dont_repeat_tonic  # 1 or 0

        scale = [get_pitch_by_number(tonic_pitch_num + (i*direction))
                 for i in range((12 * n_octaves) + repeat_tonic)]

        self.extend(scale)

    def get_triad(self, mode, index=0, degree=None):
        """Returns an array with notes from a scale's triad.

        Args:
            mode (str): Mode of the scale (eg. 'major' or 'minor')
            index (int): Triad index (eg.: 0 for 1st degree triad.)
            degree (int): Degree of the scale. If provided, overrides the
                `index` argument. (eg.: `1` for the 1st degree triad.)
        Returns:
            A list with three pitches (str), one for each note of the triad.
        """

        tonic = self.tonic

        diatonic = DiatonicScale(tonic=tonic.note, mode=mode,
                                 octave=tonic.octave, n_octaves=2,
                                 descending=False, dont_repeat_tonic=False)

        if degree:
            index = degree - 1

        form = [0, 2, 4]

        triad = [diatonic[index+note] for note in form]

        chord = Chord(triad)

        return chord



class InvalidNote(Exception):
    pass


class InvalidOctave(Exception):
    pass


class InvalidPitch(Exception):
    # def __init__(self, message=''):
    #    super(InvalidPitch, self).__init__(message)
    pass


class InvalidSequenceElement(Exception):
    pass



"__Main__ docs are here."

from . import click

from . import DIATONIC_MODES

from .interfaces.commandline import CommandLine

from . import CHROMATIC_SHARP
from . import CHROMATIC_FLAT
from . import CHROMATIC_TYPE

from .prequestion import PREQUESTION_METHODS
from .resolution import RESOLUTION_METHODS

CTX_SETTINGS = dict(
    help_option_names=['-h', '--help'],
    max_content_width=80,
)

VALID_MODES = tuple(DIATONIC_MODES.keys())
VALID_PREQUESTION_METHODS = tuple(PREQUESTION_METHODS.keys())
VALID_RESOLUTION_METHODS = tuple(RESOLUTION_METHODS.keys())

main_epilog = """
You can use 'birdears <command> --help' to show options for a specific command.

More info at https://github.com/iacchus/birdears
"""

tonics = list(set(CHROMATIC_SHARP + CHROMATIC_FLAT))
tonics.sort()

valid_tonics = ", ".join(tonics)
valid_modes = ", ".join(VALID_MODES)
valid_prequestion_methods = ", ".join(VALID_PREQUESTION_METHODS)
valid_resolution_methods = ", ".join(VALID_RESOLUTION_METHODS)


@click.group(options_metavar='', subcommand_metavar='<command> [options]',
             epilog=main_epilog,
             context_settings=CTX_SETTINGS)
@click.option('--debug/--no-debug',
              help='Turns on debugging; instead you can set DEBUG=1.',
              default=False, envvar='DEBUG')
def cli(debug):
    """birdears ─ Functional Ear Training for Musicians!"""

    if debug:
        from .logger import logger
        from .logger import logging

        global logger

        logger.setLevel(logging.DEBUG)
        logger.debug('debug is on.')


#
# melodic interval
#


melodic_epilog = """
In this exercise birdears will play two notes, the tonic and the interval
melodically, ie., one after the other and you should reply which is the correct
distance between the two.

Valid values are as follows:

-m <mode> is one of: {valid_modes}

-t <tonic> is one of: {valid_tonics}

-p <prequestion_method> is one of: {valid_prequestion_methods}

-r <resolution_method> is one of: {valid_resolution_methods}
""".format(
    valid_modes=valid_modes,
    valid_tonics=valid_tonics,
    valid_resolution_methods=valid_resolution_methods,
    valid_prequestion_methods=valid_prequestion_methods,
)


@cli.command(options_metavar='[options]', epilog=melodic_epilog)
@click.option('-m', '--mode', type=click.Choice(VALID_MODES),
              default='major', metavar='<mode>', help="Mode of the question.")
@click.option('-t', '--tonic', type=str, default='C', metavar='<tonic>',
              help='Tonic of the question.')
@click.option('-o', '--octave', type=click.IntRange(2, 5), default=None,
              metavar='<octave>',
              help="Octave of the question.")
@click.option('-d', '--descending', is_flag=True,
              help='Whether the question interval is descending.')
@click.option('-c', '--chromatic', is_flag=True,
              help='If chosen, question has chromatic notes.')
@click.option('-n', '--n_octaves', type=click.IntRange(1, 2), default=1,
              metavar='<n max>', help='Maximum number of octaves.')
@click.option('-v', '--valid_intervals', type=str, default=CHROMATIC_TYPE,
              metavar='<1,2,..>', help='A comma-separated list without spaces\
              of valid scale degrees to be chosen for the question.')
@click.option('-q', '--user_durations', type=str, default=None,
              metavar='<1,0.5,n..>', help='A comma-separated list without\
              spaces with PRECISELY 9 floating values. Or \'n\' for default\
              duration.')
@click.option('-p', '--prequestion_method', type=str, default='tonic_only',
              metavar='<prequestion_method>',
              help='The name of a pre-question method.')
@click.option('-r', '--resolution_method', type=str, default='nearest_tonic',
              metavar='<resolution_method>',
              help='The name of a resolution method.')
def melodic(*args, **kwargs):
    """Melodic interval recognition
    """

    kwargs.update({'exercise': 'melodic'})
    CommandLine(**kwargs)


#
# harmonic interval
#


harmonic_epilog = """
In this exercise birdears will play two notes, the tonic and the interval
harmonically, ie., both on the same time and you should reply which is the
correct distance between the two.

Valid values are as follows:

-m <mode> is one of: {valid_modes}

-t <tonic> is one of: {valid_tonics}

-p <prequestion_method> is one of: {valid_prequestion_methods}

-r <resolution_method> is one of: {valid_resolution_methods}
""".format(
    valid_modes=valid_modes,
    valid_tonics=valid_tonics,
    valid_resolution_methods=valid_resolution_methods,
    valid_prequestion_methods=valid_prequestion_methods,
)


@cli.command(options_metavar='[options]', epilog=harmonic_epilog)
@click.option('-m', '--mode', metavar='<mode>', type=click.Choice(VALID_MODES),
              default='major', help="Mode of the question.")
@click.option('-t', '--tonic', type=str, default='C', metavar='<note>',
              help='Tonic of the question.')
@click.option('-o', '--octave', type=click.IntRange(2, 5), default=None,
              metavar='<octave>',
              help="Octave of the question.")
@click.option('-d', '--descending', is_flag=True,
              help='Whether the question interval is descending.')
@click.option('-c', '--chromatic', is_flag=True,
              help='If chosen, question has chromatic notes.')
@click.option('-n', '--n_octaves', type=click.IntRange(1, 2), default=1,
              metavar='<n max>', help='Maximum number of octaves.')
@click.option('-v', '--valid_intervals', type=str, default=CHROMATIC_TYPE,
              metavar='<1,2,..>', help='A comma-separated list without spaces\
              of valid scale degrees to be chosen for the question.')
@click.option('-q', '--user_durations', type=str, default=None,
              metavar='<1,0.5,n..>', help='A comma-separated list without\
              spaces with PRECISELY 9 floating values. Or \'n\' for default\
              duration.')
@click.option('-p', '--prequestion_method', type=str, default='none',
              metavar='<prequestion_method>',
              help='The name of a pre-question method.')
@click.option('-r', '--resolution_method', type=str, default='nearest_tonic',
              metavar='<resolution_method>',
              help='The name of a resolution method.')
def harmonic(*args, **kwargs):
    """Harmonic interval recognition
    """

    kwargs.update({'exercise': 'harmonic'})
    CommandLine(**kwargs)

#
# dictation
#


dictation_epilog = """
In this exercise birdears will choose some random intervals and create a
melodic dictation with them. You should reply the correct intervals of the
melodic dictation.


Valid values are as follows:

-m <mode> is one of: {valid_modes}

-t <tonic> is one of: {valid_tonics}

-p <prequestion_method> is one of: {valid_prequestion_methods}

-r <resolution_method> is one of: {valid_resolution_methods}
""".format(
    valid_modes=valid_modes,
    valid_tonics=valid_tonics,
    valid_resolution_methods=valid_resolution_methods,
    valid_prequestion_methods=valid_prequestion_methods,
)


@cli.command(options_metavar='[options]', epilog=dictation_epilog)
@click.option('-m', '--mode', metavar='<mode>', type=click.Choice(VALID_MODES),
              default='major', help="Mode of the question.")
@click.option('-i', '--max_intervals', type=click.IntRange(2, 12), default=3,
              metavar='<n max>',
              help='Max random intervals for the dictation.')
@click.option('-x', '--n_notes', type=click.IntRange(1, 20), default=4,
              metavar='<n notes>',
              help='Number of notes for the dictation.')
@click.option('-t', '--tonic', type=str, default='C', metavar='<note>',
              help='Tonic of the question.')
@click.option('-o', '--octave', type=click.IntRange(2, 5), default=4,
              metavar='<octave>',
              help="Octave of the question.")
@click.option('-d', '--descending', is_flag=True,
              help='Wether the question interval is descending.')
@click.option('-c', '--chromatic', is_flag=True,
              help='If chosen, question has chromatic notes.')
@click.option('-n', '--n_octaves', type=click.IntRange(1, 2), default=1,
              metavar='<n max>', help='Maximum number of octaves.')
@click.option('-v', '--valid_intervals', type=str, default=CHROMATIC_TYPE,
              metavar='<1,2,..>', help='A comma-separated list without spaces\
              of valid scale degrees to be chosen for the question.')
@click.option('-q', '--user_durations', type=str, default=None,
              metavar='<1,0.5,n..>', help='A comma-separated list without\
              spaces with PRECISELY 9 floating values. Or \'n\' for default\
              duration.')
@click.option('-p', '--prequestion_method', type=str,
              default='progression_i_iv_v_i', metavar='<prequestion_method>',
              help='The name of a pre-question method.')
@click.option('-r', '--resolution_method', type=str, default='repeat_only',
              metavar='<resolution_method>',
              help='The name of a resolution method.')
def dictation(*args, **kwargs):
    """Melodic dictation
    """

    kwargs.update({'exercise': 'dictation'})
    CommandLine(**kwargs)


#
# instrumental dictation
#

instrumental_epilog = """
In this exercise birdears will choose some random intervals and create a
melodic dictation with them. You should play the correct melody in you musical
instrument.

Valid values are as follows:

-m <mode> is one of: {valid_modes}

-t <tonic> is one of: {valid_tonics}

-p <prequestion_method> is one of: {valid_prequestion_methods}

-r <resolution_method> is one of: {valid_resolution_methods}
""".format(
    valid_modes=valid_modes,
    valid_tonics=valid_tonics,
    valid_resolution_methods=valid_resolution_methods,
    valid_prequestion_methods=valid_prequestion_methods,
)


@cli.command(options_metavar='[options]', epilog=instrumental_epilog)
@click.option('-m', '--mode', metavar='<mode>', type=click.Choice(VALID_MODES),
              default='major', help="Mode of the question.")
@click.option('-w', '--wait_time', type=click.IntRange(1, 60), default=7,
              metavar='<seconds>',
              help='Time in seconds for next question/repeat.')
@click.option('-u', '--n_repeats', type=click.IntRange(1, 10), default=1,
              metavar='<times>', help='Times to repeat question.')
@click.option('-i', '--max_intervals', type=click.IntRange(2, 12), default=3,
              metavar='<n max>',
              help='Max random intervals for the dictation.')
@click.option('-x', '--n_notes', type=click.IntRange(1, 20), default=4,
              metavar='<n notes>',
              help='Number of notes for the dictation.')
@click.option('-t', '--tonic', type=str, default='C', metavar='<note>',
              help='Tonic of the question.')
@click.option('-o', '--octave', type=click.IntRange(2, 5), default=4,
              metavar='<octave>', help="Octave of the question.")
@click.option('-d', '--descending', is_flag=True,
              help='Wether the question interval is descending.')
@click.option('-c', '--chromatic', is_flag=True,
              help='If chosen, question has chromatic notes.')
@click.option('-n', '--n_octaves', type=click.IntRange(1, 2), default=1,
              metavar='<n max>', help='Maximum number of octaves.')
@click.option('-v', '--valid_intervals', type=str, default=CHROMATIC_TYPE,
              metavar='<1,2,..>', help='A comma-separated list without spaces\
              of valid scale degrees to be chosen for the question.')
@click.option('-q', '--user_durations', type=str, default=None,
              metavar='<1,0.5,n..>', help='A comma-separated list without\
              spaces with PRECISELY 9 floating values. Or \'n\' for default\
              duration.')
@click.option('-p', '--prequestion_method', type=str,
              default='progression_i_iv_v_i', metavar='<prequestion_method>',
              help='The name of a pre-question method.')
@click.option('-r', '--resolution_method', type=str, default='repeat_only',
              metavar='<resolution_method>',
              help='The name of a resolution method.')
def instrumental(*args, **kwargs):
    """Instrumental melodic time-based dictation
    """

    kwargs.update({'exercise': 'instrumental'})
    CommandLine(**kwargs)


notename_epilog = """
In this exercise birdears will play two notes, the tonic and the interval
melodically, ie., one after the other and you should reply which is the correct
name of the second note.

Valid values are as follows:

-m <mode> is one of: {valid_modes}

-t <tonic> is one of: {valid_tonics}

-p <prequestion_method> is one of: {valid_prequestion_methods}

-r <resolution_method> is one of: {valid_resolution_methods}
""".format(
    valid_modes=valid_modes,
    valid_tonics=valid_tonics,
    valid_resolution_methods=valid_resolution_methods,
    valid_prequestion_methods=valid_prequestion_methods,
)


@cli.command(options_metavar='[options]', epilog=notename_epilog)
@click.option('-m', '--mode', type=click.Choice(VALID_MODES),
              default='major', metavar='<mode>', help="Mode of the question.")
@click.option('-t', '--tonic', type=str, default='C', metavar='<tonic>',
              help='Tonic of the question.')
@click.option('-o', '--octave', type=click.IntRange(2, 5), default=4,
              metavar='<octave>',
              help="Octave of the question.")
@click.option('-d', '--descending', is_flag=True,
              help='Whether the question interval is descending.')
@click.option('-c', '--chromatic', is_flag=True,
              help='If chosen, question has chromatic notes.')
@click.option('-n', '--n_octaves', type=click.IntRange(1, 2), default=1,
              metavar='<n max>', help='Maximum number of octaves.')
@click.option('-v', '--valid_intervals', type=str, default=CHROMATIC_TYPE,
              metavar='<1,2,..>', help='A comma-separated list without spaces\
              of valid scale degrees to be chosen for the question.')
@click.option('-q', '--user_durations', type=str, default=None,
              metavar='<1,0.5,n..>', help='A comma-separated list without\
              spaces with PRECISLY 9 floating values. Or \'n\' for default\
              duration.')
@click.option('-p', '--prequestion_method', type=str, default='tonic_only',
              metavar='<prequestion_method>',
              help='The name of a pre-question method.')
@click.option('-r', '--resolution_method', type=str, default='nearest_tonic',
              metavar='<resolution_method>',
              help='The name of a resolution method.')
def notename(*args, **kwargs):
    """Note name by intervaç recognition
    """

    kwargs.update({'exercise': 'notename'})
    CommandLine(**kwargs)


#
# birdear's "load"
#


load_epilog = """
Loads exercise from file.
"""


@cli.command(options_metavar='', epilog=load_epilog)
@click.argument('filename', type=click.File(), metavar='<filename>')
def load(filename, *args, **kwargs):
    """Loads exercise from .toml config file <filename>.
    """

    from .toml import toml

    toml_file_str = filename.read()
    config_dict = toml.loads(toml_file_str)

    CommandLine(**config_dict)


if __name__ == "__main__":

    cli()



from .logger import log_event

from threading import Thread
import subprocess
import time

from .scale import ChromaticScale

from .note_and_pitch import Pitch
from .note_and_pitch import Chord

from .exception import InvalidSequenceElement

SEQUENCE_THREAD = None


class Sequence(list):
    """Register a Sequence of notes and/or chords.

    Attributes:
        elements (array_type): List of notes (strings) ou chords (list of
            strings) in this Sequence.
    """

    @log_event
    def __init__(self, elements=[], duration=2, delay=1.5, pos_delay=1):
        """Inits the Sequence with an array and sets the default times for
            playing / pausing the elements.

        Args:
            elements (array_type): List of elements in this sequence.
                (Pitch'es and/or Chord's)
            duration (float): Default playing time for each element in the
                sequence.
            delay (float): Default waiting time to play the next element
                in the sequence.
            pos_delay (float): Waiting time after playing the last element
                in the sequence.
        """

        if not all(isinstance(element, (Pitch, Chord))
                   for element in elements):
            raise InvalidSequenceElement

        super(Sequence, self).__init__(elements)

        self.duration = duration
        self.delay = delay
        self.pos_delay = pos_delay

    @log_event
    def play(self):
        global SEQUENCE_THREAD

        if hasattr(SEQUENCE_THREAD, 'is_alive') and SEQUENCE_THREAD.is_alive():
            try:
                SEQUENCE_THREAD.join()
            except KeyboardInterrupt:
                print('Ctrl+C')
                exit(0)

        # TODO: later we should passa callback and end_callback here so the
        # thread can talk to user interfaces, cli/tui/gui etc
        SEQUENCE_THREAD = Thread(target=self.async_play)
        SEQUENCE_THREAD.start()

        # FIXME: is this really needed? it is a global
        return SEQUENCE_THREAD

    @log_event
    def async_play(self):
        """Plays the Sequence elements of notes and/or chords and wait for
        `Sequence.pos_delay` seconds.
        """

        for index, element in enumerate(self):

            # is the current element to be played the last of the sequence?
            # because if it is the last, we will supress it's  playing delay
            # and will use the Sequence.pos_delay
            is_last = (index == len(self) - 1)

            if type(element) == Pitch:
                self._play_note(element, last_element=is_last)
            elif type(element) == Chord:
                self._play_chord(element, last_element=is_last)
            else:
                raise InvalidSequenceElement

            # TODO we should later get the element information and pass via a
            # dict to Sequence._async_play()'s callback so it can inform the
            # user interfaces on the status of the element current being played

        if self.pos_delay:
            self._wait(self.pos_delay)

    # FIXME: implement octave here:
    def make_chord_progression(self, tonic, mode, degrees):
        """Appends triad chord(s) to the Sequence.

        Args:
            tonic (str): Tonic note of the scale.
            mode (str): Mode of the scale from which build the triads upon.
            degrees (array_type): List with integers represending the degrees
                of each triad.
        """

        scale = ChromaticScale(tonic=tonic)

        for degree in degrees:
            chord = scale.get_triad(mode=mode, degree=degree)
            self.append(chord)

    @log_event
    def _play_note(self, pitch, last_element=False):
        """Plays a note.

        Args:
            note (Pitch): The note and octave to be played. Eg.: 'C4'
            duration (float): Duration of the note in seconds.
            delay (float): Delay after the note in seconds.
        """

        # requires sox to be installed

        duration = pitch.duration or self.duration
        delay = pitch.delay or self.delay

        # from sox manual: fade [type] fade-in-length [stop-position(=)
        # [fade-out-length]]

        # FIXME: this is experimental, revert to the old code if it is the case
        command = (
            "play -V1 -qn synth {duration} pluck {note}"
            " fade l 0 {duration} {duration} reverb"
        ).format(note=str(pitch), duration=duration)

        subprocess.Popen(command.split())

        if not last_element:
            self._wait(delay)

    @log_event
    def _play_chord(self, chord, last_element=False):
        """Plays a chord.

        Args:
            chord (Chord): An array of pitches (notes and octaves)
                to be played, representing a chord. Eg.: ['C4', 'Eb4', 'G5']
            duration (float): Duration of the chord in seconds.
            delay (float): Delay after the chord in seconds.
        """

        duration = chord.duration or self.duration
        delay = chord.delay or self.delay

        chord_plucks = str()
        for note in chord:
            chord_plucks += " pluck {} ".format(note)

        # FIXME: this is experimental, revert to the old code if it is the case
        command = (
            "play -V1 -qn synth {duration} {chord}"
            " fade l 0 {duration} {duration} reverb"
        ).format(chord=chord_plucks, duration=duration)

        subprocess.Popen(command.split())

        if not last_element:
            self._wait(delay)

    def _wait(self, seconds):
        """Waits, ie., stops execution for some time.

        Args:
            seconds (float): Seconds to wait.
        """

        time.sleep(seconds)
